################################################################################
# vax/__init__.py
################################################################################
"""PDS Ring-Moon Systems Node, SETI Institute

Functions to convert between VAX single- and double-precision floats and IEEE floats.
Conversions to/from VAX single precision are supported. However, only conversion from VAX
doubles to IEEE doubles is supported, not the reverse.

This module continues to support Python 2 in addition to Python 3.

Conversions to/from VAX single precision is supported. However, only conversions from VAX
doubles to IEEE doubles are supported, not the reverse.
"""

__all__ = ['from_vax32', 'to_vax32_bytes', 'to_vax32', 'from_vax64']

import numpy as np
import sys

try:
    from ._version import __version__
except ImportError:
    __version__ = 'Version unspecified'

_PYTHON2 = sys.version_info.major <= 2


def from_vax32(data):
    """Return equivalent single-precision IEEE value for VAX representation.

    Given a byte string, array, or array-like (something that can be converted to a numpy
    array), interpet it as VAX float32 or complex64 values and return the equivalent
    single-precision IEEE value(s).

    Args:
        data (bytes or bytearray or memoryview or str or numpy array-like):
            The input data.

            If the input is an array, the shape of that array is preserved except for the
            last axis, which may be modified to account for the new itemsize.

    Returns:
        np.array or np.float32 or np.complex64:
            The interpreted IEEE value. If the input array is complex, the returned array
            will have dtype "<c8"; otherwise, it will have dtype "<f4".
    """

    # Convert a string to bytes; also handle a Python 2 buffer
    if _PYTHON2:
        if isinstance(data, (str, buffer)):     # pragma: no cover  # noqa: F821
            data = bytes(data)
    else:
        if isinstance(data, str):
            data = bytes(data, encoding='latin8')

    # Convert the object to a NumPy array with an even number of 2-byte elements
    if isinstance(data, (bytes, bytearray, memoryview)):
        nbytes = data.nbytes if isinstance(data, memoryview) else len(data)
        if nbytes % 4 != 0:
            raise ValueError('data size is not a multiple of 4 bytes')

        array = np.frombuffer(data, dtype='<f4')
        scalar = (nbytes == 4)          # True to convert to scalar at the end
        shapeless = False
        newshape = (nbytes // 4,)       # array shape after conversion
        dtype = '<f4'

    else:
        scalar = np.isscalar(data)      # True to return a scalar
        array = np.asarray(data, order='C')
        shapeless = array.shape == ()   # True to convert back to shape ()

        array = np.atleast_1d(array)    # needed for the view to work below

        # Validate array and array-like data types; convert to LSB
        if isinstance(data, np.ndarray):
            key = array.dtype.kind + str(array.dtype.itemsize)
            if key not in {'f4', 'c8', 'u1', 'u2', 'u4', 'i1', 'i2', 'i4'}:
                raise ValueError('invalid data type for 4-byte array input: '
                                 + str(array.dtype))
            dtype = '<c8' if key == 'c8' else '<f4'

        else:
            # Conversion of array-like produces arrays with dtype "f8" or "c16"
            if array.dtype.kind == 'c':
                array = np.asarray(array, dtype='<c8')
                dtype = '<c8'
            elif array.dtype.kind in 'uif':
                array = np.asarray(array, dtype='<' + array.dtype.kind + '4')
                dtype = '<f4'
            else:
                raise ValueError('invalid data type for 4-byte array-like '
                                 'input: ' + str(array.dtype))

        # Determine array shape after conversion
        if array.itemsize in (1, 2):
            if (array.shape[-1] * array.itemsize) % 4 != 0:
                raise ValueError('last axis size is not a multiple of 4 bytes')

            last_axis = (array.shape[-1] * array.itemsize) // 4
            if last_axis == 1:
                newshape = array.shape[:-1]
            else:
                newshape = array.shape[:-1] + (last_axis,)

        else:
            newshape = array.shape

    itemsize = 8 if dtype == '<c8' else 4

    # Convert...

    # Swap pairs of bytes within words to put everything in LSB order (where the
    # sign has the highest memory address).
    #         |31              |15            |1
    # Before: mmmmmm_m1_mmmmmm seeeeeeeemm_m0_m
    # After:  seeeeeeeemm_m0_m mmmmmm_m1_mmmmmm
    # IEEE:   seeeeeeeemm_m0_m mmmmmmm_m1_mmmmm
    pairs = array.view(dtype='u2')
    pairs = pairs.reshape(-1, 2)
    swapped = pairs[:, ::-1].copy()
    swapped = swapped.reshape(-1, itemsize//2)
    # The sign, exponent, and mantissa are now aligned with IEEE layout

    # Correct for the different biases of the exponent
    ieee = swapped.view(dtype=dtype) / 4.

    if scalar:
        return ieee[0, 0]                # current shape is (1,1)
    elif shapeless:
        return ieee.reshape(())
    else:
        return ieee.reshape(newshape)


def to_vax32_bytes(array):
    """Return equivalent VAX representation for value(s) as bytes.

    Convert this number, array, or array-like into a byte string containing the binary
    representation of the equivalent VAX float32 or complex64 value(s).

    Args:
        array (numpy array-like): The input data.

    Returns:
        bytes: The VAX representation for the value(s).
    """

    # Make array contiguous, with C index order, containing 4-byte IEEE floats
    dtype = '<c8' if np.iscomplexobj(array) else '<f4'
    array = np.asarray(array, dtype=dtype, order='C')

    # Conversion involves multiplication by 4 and then a pairwise byte swap
    paired_view = (4. * np.atleast_1d(array)).view('u2')
    paired_view = paired_view.reshape(-1, 2)
    swapped = paired_view[:, ::-1].copy()

    return swapped.tobytes()


def to_vax32(array):
    """Return equivalent VAX representation for value(s) as numpy array.

    Convert this number, array, or array-like into an array of VAX float32 or complex64
    values with the same shape.

    Args:
        array (numpy array-like): The input data

    Returns:
        np.array: The VAX representation of the value(s) stored in a numpy array.

            If the input is a scalar, the returned object is an array of shape ().

            If the input is complex, the returned array will be of dtype "<c8"; otherwise,
            it will be of dtype "<f4".

            Note that this object will not be usable for arithmetic operations in its
            returned form.
    """

    # Make array contiguous, with C index order, containing 4-byte IEEE floats
    dtype = '<c8' if np.iscomplexobj(array) else '<f4'
    scalar = np.isscalar(array)
    array = np.asarray(array, dtype=dtype, order='C')

    # Construct array from converted buffer
    buffer = to_vax32_bytes(array)
    result = np.frombuffer(buffer, dtype=dtype).reshape(array.shape)

    if scalar:
        return result[()]
    else:
        return result


################################################################################
# 64-bit support: read-only
################################################################################

def from_vax64(data):
    """Return equivalent double-precision IEEE value for VAX representation.

    Given a byte string, array, or array-like (something that can be converted to a numpy
    array), interpet it as VAX float64 or complex128 values and return the equivalent
    double-precision IEEE value(s).

    Args:
        data (bytes or bytearray or memoryview or str or numpy array-like):
            The input data.

            If the input is an array, the shape of that array is preserved except for the
            last axis, which may be modified to account for the new itemsize.

    Returns:
        np.array or np.float64 or np.complex128:
            The interpreted IEEE value. If the input array is complex, the returned array
            will have dtype "<c16"; otherwise, it will have dtype "<f8".
    """

    # Convert a string to bytes; also handle a Python 2 buffer
    if _PYTHON2:
        if isinstance(data, (str, buffer)):     # pragma: no cover  # noqa: F821
            data = bytes(data)
    else:
        if isinstance(data, str):
            data = bytes(data, encoding='latin8')

    # Convert the object to a NumPy array
    if isinstance(data, (bytes, bytearray, memoryview)):
        nbytes = data.nbytes if isinstance(data, memoryview) else len(data)
        if nbytes % 8 != 0:
            raise ValueError('data size is not a multiple of 8 bytes')

        array = np.frombuffer(data)
        scalar = (nbytes == 8)          # True to convert to scalar at the end
        shapeless = False
        newshape = (nbytes // 8,)       # array shape after conversion
        dtype = '<f8'

    else:
        scalar = np.isscalar(data)      # True to return a scalar
        array = np.asarray(data, order='C')
        shapeless = array.shape == ()   # True to convert back to shape ()

        array = np.atleast_1d(array)    # needed for the view to work below

        # Validate array and array-like data types
        key = array.dtype.kind + str(array.dtype.itemsize)
        if key not in {'f8', 'c16', 'u1', 'u2', 'u4', 'u8',
                                    'i1', 'i2', 'i4', 'i8'}:
            raise ValueError('invalid data type for 8-byte array input: '
                             + str(array.dtype))
        dtype = '<c16' if key[0] == 'c' else '<f8'

        # Determine array shape after conversion
        if array.itemsize <= 4:
            if (array.shape[-1] * array.itemsize) % 8 != 0:
                raise ValueError('last axis size is not a multiple of 8 bytes')

            last_axis = (array.shape[-1] * array.itemsize) // 8
            if last_axis == 1:
                newshape = array.shape[:-1]
            else:
                newshape = array.shape[:-1] + (last_axis,)

        else:
            newshape = array.shape

    itemsize = 16 if dtype == '<c16' else 8

    # Convert...

    #             |31              |15            |1
    # D-floating: mmmmmm_m1_mmmmmm seeeeeeeemm_m0_m
    #             mmmmmm_m3_mmmmmm mmmmmm_m2_mmmmmm
    # IEEE:       seeeeeeeeeee_m0_ mmmmmmm_m1_mmmmm
    #             mmmmmm_m2_mmmmmm mmmmmmm_m3_mmmmm

    # Pairwise swap puts bytes in standard MSB order, with the sign in the byte
    # having the lowest memory address
    pairs = array.view(dtype='uint8')
    pairs = pairs.reshape(-1, 2)
    swapped = pairs[:, ::-1].copy()
    # Now we can treat each value as an 8-byte item in MSB order

    # IEEE has three extra bits of exponent, so we need to shift these values to
    # align the fields properly
    vals = swapped.reshape(-1, 8).view('>i8')
    vals = (vals + 4) >> 3  # add 4 first so rounding is not always toward zero
    # Sign, exponent, and mantissa are now positioned properly. However, the
    # exponents are wrong.

    # Correct the exponent's biases
    mask = (vals >= 0)
    vals[ mask] += 0x37e0000000000000  # noqa: E201
    vals[~mask] -= 0x3820000000000000

    vals = vals.reshape(-1, itemsize//8)
    ieee = vals.view(dtype)

    if scalar:
        return ieee[0, 0]                # current shape is (1,1)
    elif shapeless:
        return ieee.reshape(())
    else:
        return ieee.reshape(newshape)

################################################################################



################################################################################
# tests/test_vax.py
################################################################################

import numpy as np
import os
import unittest
import sys
from vax import from_vax32, to_vax32, from_vax64


PYTHON2 = sys.version_info.major <= 2


class Test_Vax(unittest.TestCase):

    def test_vax32(self):

        np.random.seed(4744)

        BIGINT = 2**24      # all conversions should be good to at least 24 bits
        SCALE = 1. / BIGINT
        EXPMIN = -125
        EXPMAX = 127

        # Single-value inversion tests
        reals = []
        complexes = []
        for k in range(1000):
            mantissa = np.random.randint(-BIGINT, BIGINT) * SCALE
            exponent = np.random.randint(EXPMIN, EXPMAX)
            real_ = mantissa * 2.**exponent
            reals.append(real_)

            mantissa = np.random.randint(-BIGINT, BIGINT) * SCALE
            exponent = np.random.randint(EXPMIN, EXPMAX)
            comp = complex(real_, mantissa * 2.**exponent)
            complexes.append(comp)

            for ieee, type_, dtype in [(real_, np.float32,   np.dtype('<f4')),
                                       (comp,  np.complex64, np.dtype('<c8'))]:

                # Scalar
                result = from_vax32(to_vax32(ieee))
                self.assertEqual(result, ieee)
                self.assertTrue(np.isscalar(result))
                self.assertIsInstance(result, type_)

                # Shapeless array
                arg = np.array(ieee, dtype=dtype)
                result = from_vax32(to_vax32(arg))
                self.assertEqual(result, arg)
                self.assertFalse(np.isscalar(result))
                self.assertIsInstance(result, np.ndarray)
                self.assertTrue(result.shape == ())
                self.assertTrue(result.dtype == dtype)

                # Array of shape (1,)
                arg = np.array([ieee], dtype=dtype)
                result = from_vax32(to_vax32(arg))
                self.assertTrue(np.all(result == arg))
                self.assertFalse(np.isscalar(result))
                self.assertIsInstance(result, np.ndarray)
                self.assertTrue(result.shape == (1,))
                self.assertTrue(result.dtype == dtype)

                # Array of shape (1,1)
                arg = np.array([[ieee]], dtype=dtype)
                result = from_vax32(to_vax32(arg))
                self.assertTrue(np.all(result == arg))
                self.assertFalse(np.isscalar(result))
                self.assertIsInstance(result, np.ndarray)
                self.assertTrue(result.shape == (1,1))
                self.assertTrue(result.dtype == dtype)

                # Array-like of shape (1,)
                result = from_vax32(to_vax32([ieee]))
                self.assertTrue(np.all(result == arg))
                self.assertFalse(np.isscalar(result))
                self.assertIsInstance(result, np.ndarray)
                self.assertTrue(result.shape == (1,))
                self.assertTrue(result.dtype == dtype)

                # Array-like of shape (1,1)
                result = from_vax32(to_vax32([[ieee]]))
                self.assertTrue(np.all(result == arg))
                self.assertFalse(np.isscalar(result))
                self.assertIsInstance(result, np.ndarray)
                self.assertTrue(result.shape == (1,1))
                self.assertTrue(result.dtype == dtype)

        # Array-like inversion tests
        results = from_vax32(to_vax32(reals))
        self.assertTrue(np.all(results == reals))

        results = from_vax32(to_vax32(complexes))
        self.assertTrue(np.all(results == complexes))

        test = [[reals[   :125], reals[125:250], reals[250:375], reals[375:500]],
                [reals[500:625], reals[625:750], reals[750:875], reals[875:]]]
        results = from_vax32(to_vax32(test))
        self.assertTrue(np.all(results == test))
        self.assertEqual(results.shape, (2,4,125))

        test = list(range(10))
        self.assertEqual(test, list(from_vax32(to_vax32(test))))

        # Array inversion tests
        for k in range(10):
            for shape in [(7,), (7,7), (7,7,7), (4,1), (4,2), (4,3),
                          (3,), (3,1), (3,1,1), (1,3), (1,3,1), (1,1,3),
                          (1,3,1,1,1,1)]:
                mantissa = np.random.randint(-BIGINT, BIGINT, size=shape) * SCALE
                exponent = np.random.randint(EXPMIN, EXPMAX, size=shape)
                real_ = mantissa * 2.**exponent

                cshape = shape + (2,)
                mantissa = np.random.randint(-BIGINT, BIGINT, size=cshape) * SCALE
                exponent = np.random.randint(EXPMIN, EXPMAX, size=cshape)
                temp = mantissa * 2.**exponent
                comp = temp.view('complex').reshape(shape)

                for ieee, dtype in [(real_, np.dtype('<f4')),
                                    (comp,  np.dtype('<c8'))]:
                    result = from_vax32(to_vax32(ieee))
                    self.assertTrue(isinstance(result, np.ndarray))
                    self.assertTrue(result.shape == ieee.shape)
                    self.assertTrue(result.dtype == dtype)
                    self.assertTrue(np.all(result == ieee))

        # Single-value buffer, memoryview, bytes, bytearray, str
        for k in range(100):
            mantissa = np.random.randint(-BIGINT, BIGINT) * SCALE
            exponent = np.random.randint(EXPMIN, EXPMAX)
            ieee = mantissa * 2.**exponent

            # buffer (Python 2) or memoryview (Python 3)
            vax32 = to_vax32(ieee)
            result = from_vax32(vax32.data)
            self.assertTrue(np.isscalar(result))
            self.assertEqual(result, ieee)

            # bytes
            result = from_vax32(bytes(vax32.data))
            self.assertTrue(np.isscalar(result))
            self.assertEqual(result, ieee)

            # bytearray
            result = from_vax32(bytearray(vax32.data))
            self.assertTrue(np.isscalar(result))
            self.assertEqual(result, ieee)

            # string
            if PYTHON2:
                result = from_vax32(str(vax32.data))    # pragma: no cover
            else:
                result = from_vax32(str(vax32.data, encoding='latin8'))
            self.assertTrue(np.isscalar(result))
            self.assertEqual(result, ieee)

        # Multiple-value buffer, memoryview, bytes, bytearray, str
        for k in range(100):
            size = np.random.randint(2, 21)
            mantissa = np.random.randint(-BIGINT, BIGINT, size=size) * SCALE
            exponent = np.random.randint(EXPMIN, EXPMAX, size=size)
            ieee = (mantissa * 2.**exponent).astype('<f4')

            # buffer (Python 2) or memoryview (Python 3)
            vax32 = to_vax32(ieee)
            result = from_vax32(vax32.data)
            self.assertTrue(result.shape == ieee.shape)
            self.assertTrue(np.all(result == ieee))

            # bytes
            result = from_vax32(bytes(vax32.data))
            self.assertTrue(result.shape == ieee.shape)
            self.assertTrue(np.all(result == ieee))

            # bytearray
            result = from_vax32(bytearray(vax32.data))
            self.assertTrue(result.shape == ieee.shape)
            self.assertTrue(np.all(result == ieee))

            # string
            if PYTHON2:
                result = from_vax32(str(vax32.data))    # pragma: no cover
            else:
                result = from_vax32(str(vax32.data, encoding='latin8'))
            self.assertTrue(result.shape == ieee.shape)
            self.assertTrue(np.all(result == ieee))

        # Try some real-world Vax data from
        #   VGISS_5xxx/VGISS_5214/CALIB/MIPL/VGRSCF.DAT
        #
        # array = np.fromfile('.../VGRSCF.DAT', dtype='uint8')[780:]

        uints = np.array([
          104,  64,  39,  49,  96,  64, 156, 196,  60,  64,   8, 172, 128,
           64,   0,   0, 125,  64,  27,  47, 128,  64,   0,   0, 131,  64,
           10, 215,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0, 110,  64,   4,  86,  77,  64,
          242, 210, 128,  64,   0,   0, 134,  64,  25,   4, 110,  64,   4,
           86, 132,  64,  88,  57, 132,  64,  88,  57,   8,  65, 176, 114,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,  32,  54,  47,  49,  53,  47,  56,  53,
          128,  64,   0,   0, 136,  64,  57, 180, 130,  64, 229, 208, 101,
           64, 203, 161, 128,  64,   0,   0, 128,  64,  78,  98, 128,  64,
            0,   0, 125,  64,  27,  47,  86,  65, 123,  20, 100,  65, 166,
          155,  90,  65, 131, 192,  64,  65,   0,   0, 128,  64,   0,   0,
           86,  65,  82, 184,  86,  65, 123,  20,  83,  65,  70, 182,  88,
           66,  53,  94, 103,  66, 111,  18,  93,  66, 184,  30,  66,  66,
          123,  20, 128,  64,   0,   0,  89,  66,  12,   2,  88,  66,  53,
           94,  85,  66, 231, 251,  88,  66,  53,  94, 103,  66, 111,  18,
           93,  66, 184,  30,  66,  66, 123,  20, 128,  64,   0,   0,  89,
           66,  12,   2,  88,  66,  53,  94,  85,  66, 231, 251, 105,  64,
           94, 186,  98,  64, 211,  77, 128,  64,   0,   0, 118,  64,  25,
            4, 105,  64,  94, 186, 129,  64, 252, 169, 129,  64, 252, 169,
           73,  65, 252, 169,  67,  65, 188, 116,  61,  65, 125,  63,  86,
           65, 123,  20,  77,  65, 143, 194,  67,  65, 188, 116,  88,  65,
          254, 212,  88,  65, 254, 212,  40,  66, 215, 163,  69,  66,  55,
          137,  63,  66, 150,  67,  88,  66,  53,  94,  79,  66, 231, 251,
           69,  66,  55, 137,  91,  66,   2,  43,  91,  66,   2,  43,  42,
           67, 170, 113,  69,  66,  55, 137,  63,  66, 150,  67,  88,  66,
           53,  94,  79,  66, 231, 251,  69,  66,  55, 137,  91,  66,   2,
           43,  91,  66,   2,  43,  42,  67, 170, 113, 128,  64,   0,   0,
          104,  64,  39,  49,  96,  64, 156, 196,  60,  64,   8, 172, 128,
           64,   0,   0, 125,  64,  27,  47, 128,  64,   0,   0, 131,  64,
           10, 215,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0, 110,  64,   4,  86,  77,  64,
          242, 210, 128,  64,   0,   0, 134,  64,  25,   4, 110,  64,   4,
           86, 132,  64,  88,  57, 132,  64,  88,  57,   8,  65, 176, 114,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
            0,   0,   0,   0,   0,  32,  50,  47,  48,  50,  47,  56,  54],
            dtype='uint8')

        ieee = from_vax32(uints)
        truth = np.array([1.   , 1.068, 1.022, 0.897, 1.   , 1.003, 1.   , 0.989])
        self.assertTrue(np.all(truth.astype('<f4') == ieee[65:73]))

        truth = np.array([3.345, 3.572, 3.418, 3.   , 1.   , 3.355, 3.345, 3.308])
        self.assertTrue(np.all(truth.astype('<f4') == ieee[73:81]))

        truth = np.array([13.523, 14.442, 13.82 , 12.13 ,  1.   , 13.563, 13.523, 13.374])
        self.assertTrue(np.all(truth.astype('<f4') == ieee[81:89]))
        self.assertTrue(np.all(truth.astype('<f4') == ieee[89:97]))

        ieee = from_vax32(uints.reshape(39,20))
        self.assertEqual(ieee.shape, (39,5))

        ieee = from_vax32(uints.reshape(39,5,4))
        self.assertEqual(ieee.shape, (39,5))

        self.assertRaises(ValueError, from_vax32, uints[:9])  # not a multiple of 4

        # Errors
        self.assertRaises(ValueError, from_vax32, b'1')
        self.assertRaises(ValueError, from_vax32, np.arange(10, dtype='f8'))
        self.assertRaises(ValueError, from_vax32, np.array(['123', '456']))
        self.assertRaises(ValueError, from_vax32, ['123', '456'])

        # Reading from a file...
        vax_dir = os.path.split(sys.modules['vax'].__file__)[0]
        parent = os.path.split(vax_dir)[0]
        test_file = os.path.join(parent, 'test_files', 'C3490702_GEOMA.DAT')
        data = np.fromfile(test_file, dtype='<f4')
        data = data[1536//4:]       # skip LBLSIZE
        data = from_vax32(data)
        data = data.reshape(-1,4)   # four columns

        answer_10 = np.array([
            [ 25.36      ,  25.31      ,   9.831727  ,  15.862788  ],
            [ 25.36      ,  25.31      ,   9.831727  ,  15.862788  ],
            [ 20.53      ,  85.59      ,   2.6002007 ,  60.86287   ],
            [ 20.53      ,  85.59      ,   2.6002007 ,  60.86287   ],
            [ 25.36      , 177.6       ,   3.0882928 , 133.30754   ],
            [ 25.36      , 177.6       ,   3.0882928 , 133.30754   ],
            [ 25.36      , 269.74      ,   1.6376413 , 208.92227   ],
            [ 25.36      , 269.74      ,   1.6376413 , 208.92227   ],
            [ 25.36      , 362.04      ,   0.71304405, 285.84558   ],
            [ 25.36      , 362.04      ,   0.71304405, 285.84558   ]], dtype='float32')
        self.assertTrue(np.all(answer_10 == data[:10]))

        # Tests copied from libvaxdata
        vax_f4 = np.array(
            [0x80, 0x40, 0x00, 0x00,
             0x80, 0xC0, 0x00, 0x00,
             0x60, 0x41, 0x00, 0x00,
             0x60, 0xC1, 0x00, 0x00,
             0x49, 0x41, 0xD0, 0x0F,
             0x49, 0xC1, 0xD0, 0x0F,
             0xF0, 0x7D, 0xC2, 0xBD,
             0xF0, 0xFD, 0xC2, 0xBD,
             0x08, 0x03, 0xEA, 0x1C,
             0x08, 0x83, 0xEA, 0x1C,
             0x9E, 0x40, 0x52, 0x06,
             0x9E, 0xC0, 0x52, 0x06],
            dtype='uint8').reshape(-1,4)

        ieee_f4 = np.array([
            1.000000    ,
           -1.000000    ,
            3.500000    ,
           -3.500000    ,
            3.141590    ,
           -3.141590    ,
            9.9999999E+36,
           -9.9999999E+36,
            9.9999999E-38,
           -9.9999999E-38,
            1.234568    ,
           -1.234568    ], dtype='float32')

        diff = np.abs((from_vax32(vax_f4) - ieee_f4) / ieee_f4)
        self.assertEqual(np.median(diff), 0.)
        self.assertTrue(np.all(diff < 1.e-7))
        answer = from_vax32(vax_f4)

        self.assertTrue(np.all(from_vax32(bytes(vax_f4.data)) == answer))

        if not PYTHON2:
            string = bytes(vax_f4.data).decode('latin8')
            self.assertTrue(np.all(from_vax32(string) == answer))

        ints = vax_f4.view('<i2')
        self.assertTrue(np.all(from_vax32(ints) == answer))

        ints = vax_f4.view('>i2')
        self.assertTrue(np.all(from_vax32(ints) == answer))

        ints = vax_f4.view('<i4').reshape(-1)
        self.assertTrue(np.all(from_vax32(ints) == answer))

        ints = vax_f4.view('>i4').reshape(-1)
        self.assertTrue(np.all(from_vax32(ints) == answer))

        floats = vax_f4.view('<f4').reshape(-1)
        self.assertTrue(np.all(from_vax32(floats) == answer))

        floats = vax_f4.view('>f4').reshape(-1)
        self.assertTrue(np.all(from_vax32(floats) == answer))

        comps = vax_f4.reshape(-1,8).view('<c8').reshape(-1)
        ieee_c8 = answer.view('<c8')
        self.assertTrue(np.all(from_vax32(comps) == ieee_c8))

        comps = vax_f4.reshape(-1,8).view('>c8').reshape(-1)
        self.assertTrue(np.all(from_vax32(comps) == ieee_c8))

        floats = vax_f4.view('=f4').reshape(-1)
        for k,x in enumerate(floats):
            self.assertEqual(from_vax32(x), answer[k])

        comps = vax_f4.reshape(-1,8).view('=c8').reshape(-1)
        for k,x in enumerate(comps):
            self.assertEqual(from_vax32(x), ieee_c8[k])

    def test_vax64(self):

        # Tests copied from libvaxdata
        vax_d8 = np.array(
            [0x80, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x80, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x60, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x60, 0xC1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x49, 0x41, 0xDA, 0x0F, 0x21, 0xA2, 0xBE, 0x68,
             0x49, 0xC1, 0xDA, 0x0F, 0x21, 0xA2, 0xBE, 0x68,
             0xF0, 0x7D, 0xC2, 0xBD, 0xBB, 0x1A, 0xDB, 0x48,
             0xF0, 0xFD, 0xC2, 0xBD, 0xBB, 0x1A, 0xDB, 0x48,
             0x08, 0x03, 0xEA, 0x1C, 0x54, 0x14, 0x75, 0x5C,
             0x08, 0x83, 0xEA, 0x1C, 0x54, 0x14, 0x75, 0x5C,
             0x9E, 0x40, 0x52, 0x06, 0x62, 0x14, 0xE7, 0xCE,
             0x9E, 0xC0, 0x52, 0x06, 0x62, 0x14, 0xE7, 0xCE],
            dtype='uint8').reshape(-1,8)

        ieee_f8 = np.array([
            1.000000000000000    ,
           -1.000000000000000    ,
            3.500000000000000    ,
           -3.500000000000000    ,
            3.141592653589793    ,
           -3.141592653589793    ,
            1.0000000000000000E+37,
           -1.0000000000000000E+37,
            9.9999999999999999E-38,
           -9.9999999999999999E-38,
            1.234567890123450    ,
           -1.234567890123450    ], dtype='float64')

        self.assertTrue(np.all(from_vax64(vax_d8) == ieee_f8))

        self.assertTrue(np.all(from_vax64(vax_d8.data) == ieee_f8))

        self.assertTrue(np.all(from_vax64(bytes(vax_d8.data)) == ieee_f8))

        if not PYTHON2:
            string = bytes(vax_d8.data).decode('latin8')
            self.assertTrue(np.all(from_vax64(string) == ieee_f8))

        ints = vax_d8.view('<i2')
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('>i2')
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('<i4')
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('>i4')
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('<i8').reshape(-1)
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('>i8').reshape(-1)
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        floats = vax_d8.view('<f8').reshape(-1)
        self.assertTrue(np.all(from_vax64(floats) == ieee_f8))

        floats = vax_d8.view('>f8').reshape(-1)
        self.assertTrue(np.all(from_vax64(floats) == ieee_f8))

        comps = vax_d8.reshape(-1,16).view('<c16').reshape(-1)
        ieee_c16 = ieee_f8.view('<c16')
        self.assertTrue(np.all(from_vax64(comps) == ieee_c16))

        comps = vax_d8.reshape(-1,16).view('>c16').reshape(-1)
        self.assertTrue(np.all(from_vax64(comps) == ieee_c16))

        floats = vax_d8.view('=f8').reshape(-1)
        for k,x in enumerate(floats):
            self.assertEqual(from_vax64(x), ieee_f8[k])

        for k,x in enumerate(floats):
            self.assertEqual(from_vax64(np.array(x)), np.array(ieee_f8[k]))

        comps = vax_d8.reshape(-1,16).view('=c16').reshape(-1)
        for k,x in enumerate(comps):
            self.assertEqual(from_vax64(x), ieee_c16[k])

        ints = vax_d8.view('>i2').reshape(-1,4)
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8))

        ints = vax_d8.view('>i2').reshape(-1,8)
        self.assertTrue(np.all(from_vax64(ints) == ieee_f8.reshape(-1,2)))

        # Errors
        self.assertRaises(ValueError, from_vax64, b'1')
        self.assertRaises(ValueError, from_vax64, np.array(['123', '456']))
        self.assertRaises(ValueError, from_vax64, ['123', '456'])
        self.assertRaises(ValueError, from_vax64, vax_d8.reshape(-1)[:7])
        self.assertRaises(ValueError, from_vax64, vax_d8[:,:4])
